from typing import Any

try:
    from typing import Protocol
except ImportError:  # pragma: no cover
    from typing_extensions import Protocol  # type: ignore[assignment]


class QueueProtocol(Protocol):
    def put_nowait(self, item: Any) -> None:
        ...  # pragma: no cover

    def qsize(self) -> int:
        ...  # pragma: no cover

    def get(self) -> Any:
        ...  # pragma: no cover

    def task_done(self) -> None:
        ...  # pragma: no cover



from abc import ABC, abstractmethod
from typing import List, Callable, Any, Optional

try:
    from functools import cached_property  # type: ignore[attr-defined, unused-ignore]
except ImportError:
    from cached_property import cached_property  # type: ignore[import-not-found, no-redef, import, unused-ignore]

from awaits.task import Task
from awaits.protocols.queue import QueueProtocol
from awaits.units.abstract_unit import AbstractUnit


class AbstractPool(ABC):
    def __init__(self, size: int) -> None:
        if size <= 0:
            raise ValueError('The size of the pool must be greater than zero.')

        self.size = size
        self.workers = self.create_workers()

    def __len__(self) -> int:
        return self.size

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({len(self)})'

    def __str__(self) -> str:
        return f'<{self.__class__.__name__} pool object of {len(self)} workers #{id(self)}>'

    def __getitem__(self, index: int) -> AbstractUnit:
        if not isinstance(index, int):
            raise ValueError('Key must be an integer number.')
        if index >= len(self) or index < 0:
            raise IndexError(f'The size of the pool is equal {len(self)}.')
        return self.workers[index]

    @cached_property  # type: ignore[misc, unused-ignore]
    def queue(self) -> QueueProtocol:
        return self.get_queue()

    def do(self, function: Callable[[Any], Any], *args: Any, **kwargs: Any) -> Task:
        task = Task(function, *args, **kwargs)
        self.queue.put_nowait(task)
        return task

    def create_workers(self, number_of_workers: Optional[int] = None, base_number: int = 0) -> List[AbstractUnit]:
        pool_size = self.size if number_of_workers is None else number_of_workers
        workers = []

        for number in range(base_number, pool_size):
            worker = self.create_worker(number)
            workers.append(worker)
        return workers

    @abstractmethod
    def get_queue(self) -> QueueProtocol:
        ... # pragma: no cover

    @abstractmethod
    def create_worker(self, index: int) -> AbstractUnit:
        ... # pragma: no cover



from queue import Queue

from awaits.units.abstract_unit import AbstractUnit
from awaits.units.thread_unit import ThreadUnit
from awaits.pools.abstract_pool import AbstractPool
from awaits.protocols.queue import QueueProtocol


class ThreadsPool(AbstractPool):
    def get_queue(self) -> QueueProtocol:
        return Queue()

    def create_worker(self, index: int) -> AbstractUnit:
        return ThreadUnit(self.queue, self, index)



from typing import Tuple, Dict, Callable, Union, Any, Optional


class Task:
    def __init__(self, function: Callable[[Any], Any], *args: Any, **kwargs: Any) -> None:
        self.function: Callable[[Any], Any] = function
        self.args: Tuple[Any, ...] = args
        self.kwargs: Dict[str, Any] = kwargs
        self.done: bool = False
        self.error: bool = False
        self.result: Optional[Any] = None
        self.exception: Optional[BaseException] = None

    def __call__(self) -> None:
        try:
            self.result = self.function(*(self.args), **(self.kwargs))
        except Exception as e:
            self.error = True
            self.exception = e
        self.done = True

    def __repr__(self) -> str:
        def string_wrapper(some_value: Union[str, Any]) -> str:
            return f'"{some_value}"' if isinstance(some_value, str) else repr(some_value)

        function_name = self.function.__name__
        args = ', '.join([string_wrapper(x) for x in self.args])
        kwargs = ', '.join([f'{key}={string_wrapper(value)}' for key, value in self.kwargs.items()])
        content = []
        if function_name:
            content.append(function_name)
        if args:
            content.append(args)
        if kwargs:
            content.append(kwargs)
        joined_content = ', '.join(content)
        result = f'Task({joined_content})'
        return result



import pytest

from awaits import config


def test_set_incompatible_typed_variable():
    with pytest.raises(TypeError, match='Variable "kek" has not the type int.'):
        config.set(pool_size='kek')

    with pytest.raises(TypeError, match='Variable "kek" has not one of types: int, float.'):
        config.set(delay='kek')


def test_unknown_key_name():
    with pytest.raises(KeyError, match='"kek" variable is not allowed for the awaits settings.'):
        config.set(kek='kek')



from queue import Queue
from awaits import shoot


def test_simple_shoot():
    queue = Queue()
    lst = []

    @shoot
    def function():
        for number in range(100):
            lst.append(number)
        queue.put(None)

    function()
    queue.get()

    assert len(lst) == 100



from abc import ABC, abstractmethod

from awaits.protocols.queue import QueueProtocol


class AbstractUnit(ABC):
    def __init__(self, queue: QueueProtocol, pool: 'AbstractPool', index: int) -> None:  # type: ignore[name-defined] # noqa: F821
        self.index = index
        self.queue = queue
        self.pool = pool
        self.activate()

    @abstractmethod
    def activate(self) -> None:
        ... # pragma: no cover

    @abstractmethod
    def run(self) -> None:
        ... # pragma: no cover



from typing import Dict, Tuple, Type, Union

from awaits.common_data import CommonData


class config:
    """
    Установка глобальных параметров.
    """
    allowed_settings: Dict[str, Tuple[Union[Type[int], Type[float]], ...]] = {
        'pool_size': (int, ),
        'delay': (int, float, ),
    }

    @staticmethod
    def set(**kwargs: Union[int, float]) -> None:
        new_kwargs = {}

        for key, value in kwargs.items():
            if key not in config.allowed_settings:
                raise KeyError(f'"{key}" variable is not allowed for the awaits settings.')
            allowed_types = config.allowed_settings[key]

            is_allowed = False
            for one in allowed_types:
                if isinstance(value, one):
                    is_allowed = True

            if not is_allowed:
                if len(allowed_types) == 1:
                    raise TypeError(f'Variable "{value}" has not the type {allowed_types[0].__name__}.')
                else:
                    allowed_types_in_line = ', '.join([allowed_type.__name__ for allowed_type in allowed_types])
                    raise TypeError(f'Variable "{value}" has not one of types: {allowed_types_in_line}.')

            new_kwargs[key] = value

        CommonData(**new_kwargs)



from awaits.common_data import CommonData
from awaits.threads_pools_room import ThreadsPoolsRoom


class RoomKeeper:
    def __init__(self) -> None:
        if not hasattr(self, 'room'):
            pool_size = CommonData().pool_size
            self.room = ThreadsPoolsRoom(pool_size)

    def __new__(cls) -> 'RoomKeeper':
        if not hasattr(cls, 'instance'):
            cls.instance = super(RoomKeeper, cls).__new__(cls)
        return cls.instance



from typing import Tuple, Callable, Any, Union

from awaits.errors import IncorrectUseOfTheDecoratorError


def end_of_wrappers(args: Tuple[Callable[[Any], Any], ...], wrapper: Callable[[Callable[[Any], Any]], Callable[[Any], Any]]) -> Union[Callable[[Callable[[Any], Any]], Callable[[Any], Any]], Callable[[Any], Any]]:
    """
    Определяем, как вызван декоратор - как фабрика декораторов (т. е. без позиционных аргументов) или как непосредственный декоратор.
    """
    if not len(args):
        return wrapper
    elif len(args) == 1 and callable(args[0]):
        return wrapper(args[0])
    raise IncorrectUseOfTheDecoratorError('You used the awaitable decorator incorrectly. Read the documentation.')



from awaits.task import Task


def test_run_simple_function():
    value = 777

    def function():
        return value

    task = Task(function)

    assert task.done == False
    assert task.result is None
    assert task.error == False
    assert task.exception is None

    task()

    assert task.done == True
    assert task.result == value
    assert task.error == False
    assert task.exception is None


def test_run_function_with_arguments():
    first_value = 777
    second_value = 999
    third_value = 5

    def function(a, b, c=third_value):
        return a + b + c

    task = Task(function, first_value, second_value, c=third_value)

    assert task.done == False
    assert task.result is None
    assert task.error == False
    assert task.exception is None

    task()

    assert task.done == True
    assert task.result == first_value + second_value + third_value
    assert task.error == False
    assert task.exception is None


def test_run_function_that_raise_exception():
    first_value = 777
    second_value = 999
    third_value = 5

    def function(a, b, c=third_value):
        raise ValueError

    task = Task(function, first_value, second_value, c=third_value)

    assert task.done == False
    assert task.result is None
    assert task.error == False
    assert task.exception is None

    task()

    assert task.done == True
    assert task.result == None
    assert task.error == True
    assert isinstance(task.exception, ValueError)


def test_repr_task():
    def function(a, b, c):
        return

    task = Task(function, 1, 2, c="kek")

    assert repr(task) == f'Task(function, 1, 2, c="kek")'
    assert str(task) == f'Task(function, 1, 2, c="kek")'



from typing import Union


class CommonData:
    pool_size: int = 10
    delay: Union[int, float] = 0.01

    def __init__(self, **kwargs: Union[int, float]) -> None:
        for key, value in kwargs.items():
            setattr(self.__class__, key, value)

    def __new__(cls, **kwargs: Union[int, float]) -> 'CommonData':
        if not hasattr(cls, 'instance'):
            cls.instance = super(CommonData, cls).__new__(cls)
        return cls.instance



from threading import Lock
from typing import Dict

from awaits.pools.threads_pool import ThreadsPool


class ThreadsPoolsRoom:
    def __init__(self, pool_size: int) -> None:
        self.pool_size: int = pool_size
        self.pools: Dict[str, ThreadsPool] = {}
        self.lock: Lock = Lock()

    def __getitem__(self, key: str) -> ThreadsPool:
        if not isinstance(key, str):
            raise KeyError(f'Key of the pools room must be a string, not {key.__class__.__name__}.')
        with self.lock:
            if key not in self.pools:
                self.pools[key] = ThreadsPool(self.pool_size)
            return self.pools[key]

    def __len__(self) -> int:
        return len(self.pools)



from typing import Union, Optional

from awaits.errors import IncorrectUseOfTheDecoratorError
from awaits.pools.abstract_pool import AbstractPool
from awaits.room_keeper import RoomKeeper


def get_pool_for_decorator(pool: Optional[Union[str, AbstractPool]]) -> AbstractPool:
    if pool is None:
        return RoomKeeper().room['base']
    elif isinstance(pool, str):
        return RoomKeeper().room[pool]
    elif isinstance(pool, AbstractPool):
        return pool
    raise IncorrectUseOfTheDecoratorError('You can specify the pool by name, or pass an instance of the class inherited from AbstractPool.')



from functools import wraps
from asyncio import sleep
from typing import Awaitable, Callable, Any, Optional, Union

from awaits.common_data import CommonData
from awaits.utils.get_pool_for_decorator import get_pool_for_decorator
from awaits.utils.end_of_wrappers import end_of_wrappers
from awaits.pools.abstract_pool import AbstractPool


def awaitable(*args: Callable[[Any], Any], pool: Optional[Union[str, AbstractPool]] = None, delay: Optional[Union[int, float]] = None) -> Union[Callable[[Callable[[Any], Any]], Callable[[Any], Awaitable[Any]]], Callable[[Any], Awaitable[Any]]]:
    pool = get_pool_for_decorator(pool)

    def wrapper_of_wrapper(function: Callable[[Any], Any]) -> Callable[[Any], Awaitable[Any]]:
        @wraps(function)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            sleep_time = CommonData().delay if delay is None else delay
            task = pool.do(function, *args, **kwargs)
            while not task.done:
                await sleep(sleep_time)
            if task.error:
                raise task.exception  # type: ignore[misc]
            return task.result
        return wrapper

    return end_of_wrappers(args, wrapper_of_wrapper)



from threading import Thread

from awaits.units.abstract_unit import AbstractUnit


class ThreadUnit(AbstractUnit):
    """
    Экземпляр класса соответствует одному потоку. Здесь выполняются таски.
    """
    def activate(self) -> None:
        self.thread = Thread(target=self.run, args=())
        self.thread.daemon = True
        self.thread.start()

    def run(self) -> None:
        """
        Принимаем из очереди таски и выполняем их.
        """
        while True:
            try:
                task = self.queue.get()
                task()
            finally:
                self.queue.task_done()



from asyncio import run
from threading import get_ident

import pytest

from awaits import awaitable
from awaits.errors import IncorrectUseOfTheDecoratorError
from awaits.room_keeper import RoomKeeper


def test_await_awaitable_function_without_brackets_without_arguments():
    value = 777

    @awaitable
    def function():
        return value

    assert run(function()) == value


def test_await_awaitable_function_without_brackets_without_arguments_raise_exception():
    @awaitable
    def function():
        raise ValueError

    with pytest.raises(ValueError):
        run(function())


def test_check_thread_id():
    @awaitable
    def function():
        return get_ident()

    assert run(function()) != get_ident()


def test_await_awaitable_function_without_brackets_with_arguments():
    first_value = 777
    second_value = 999
    third_value = 5
    forth_value = 17

    @awaitable
    def function(a, b, c=third_value):
        return a + b + c

    assert run(function(first_value, second_value)) == first_value + second_value + third_value
    assert run(function(first_value, second_value, c=forth_value)) == first_value + second_value + forth_value


def test_await_awaitable_function_with_empty_brackets_without_arguments():
    value = 777

    @awaitable()
    def function():
        return value

    assert run(function()) == value


def test_await_awaitable_function_with_empty_brackets_with_arguments():
    first_value = 777
    second_value = 999
    third_value = 5
    forth_value = 17

    @awaitable()
    def function(a, b, c=third_value):
        return a + b + c

    assert run(function(first_value, second_value)) == first_value + second_value + third_value
    assert run(function(first_value, second_value, c=forth_value)) == first_value + second_value + forth_value


def test_await_awaitable_function_with_brackets_and_pool_name_without_arguments():
    value = 777

    @awaitable(pool='kek')
    def function():
        return value

    assert run(function()) == value


def test_await_awaitable_function_with_brackets_and_wrong_pool_object_without_arguments():
    value = 777

    with pytest.raises(IncorrectUseOfTheDecoratorError):
        @awaitable(pool=5)
        def function():
            return value


def test_await_awaitable_function_with_brackets_and_pool_object_without_arguments():
    value = 777
    pool = RoomKeeper().room['test_await_awaitable_function_with_brackets_and_pool_object_without_arguments']

    @awaitable(pool=pool)
    def function():
        return value

    assert run(function()) == value


def test_await_awaitable_function_with_brackets_and_pool_object_without_arguments():
    value = 777
    pool = RoomKeeper().room['test_await_awaitable_function_with_brackets_and_pool_object_without_arguments']

    @awaitable(pool=pool)
    def function():
        return value

    assert run(function()) == value



from functools import wraps
from typing import Callable, Union, Optional, Any

from awaits.utils.get_pool_for_decorator import get_pool_for_decorator
from awaits.utils.end_of_wrappers import end_of_wrappers
from awaits.pools.abstract_pool import AbstractPool
from awaits.task import Task


def shoot(*args: Callable[[Any], Any], pool: Optional[Union[str, AbstractPool]] = None) -> Union[Callable[[Callable[[Any], Any]], Callable[[Any], Any]], Callable[[Any], Any]]:
    pool = get_pool_for_decorator(pool)

    def wrapper_of_wrapper(function: Callable[[Any], Any]) -> Callable[[Any], Task]:
        @wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Task:
            task = pool.do(function, *args, **kwargs)
            return task

        return wrapper

    return end_of_wrappers(args, wrapper_of_wrapper)
